import time
from SimulatorFramework import TileAssembly
import RobinsonTiles
from SimulatorFramework import TileType, Tile
from SimulatorFramework import SimUtils
import os
import ast
import random
import itertools
from test.test_deque import BIG


class RobinsonAssembly(TileAssembly.Assembly):
    def __init__(self, tileWidths, tileHeights):
        TileAssembly.Assembly.__init__(self, tileWidths, tileHeights)

        self.growthPattern = []
        self.growthPatternLoc = 0
        self.growthTileLocations = [(0,0)]
        self.tileTypesUsed = set()
        self.inputDict = dict()
        self.twoNbrsToPlace = []
        self.subsetOfNbrsToPlace = []

    def ClearSpace(self):
        TileAssembly.Assembly.ClearSpace(self)
        self.growthPatternLoc = 0
        self.growthTileLocations = [(0,0)]
        self.tileTypesUSed = set()
        self.inputDict = dict()
        self.twoNbrsToPlace = []
        self.subsetOfNbrsToPlace = []
        
    def AddToFrontierIfNecessary(self, coords):
        if RobinsonTiles.RobinsonFrontierLocation(coords, None) in self.frontier:
            return None
        fittingNbrSet = self.GetFittingTilesForLoc(coords)
        if fittingNbrSet is None:
            return None
        elif len(fittingNbrSet) == 1:
            fLoc = RobinsonTiles.RobinsonFrontierLocation(coords, fittingNbrSet)
            self.frontier.discard(fLoc)
            self.frontier.add(fLoc)
            if fLoc in self.disabledFrontier:
                self.ClearLoc(coords)
                return None
            else:
                self.SetLoc(coords, TileAssembly.FRONTIER_LOC)
                return fLoc
        #elif len(fittingNbrSet) > 1 and fittingNbrSet.issubset(self.growthTileTypes):
        elif (len(fittingNbrSet) > 1) and (self.IsSetOfGrowthTiles(fittingNbrSet) == True):
            fLoc = RobinsonTiles.RobinsonFrontierLocation(coords, fittingNbrSet)
            self.frontier.add(fLoc)
            if fLoc in self.disabledFrontier:
                SimUtils.PrintDebug("Adding a disabled growth tile frontier location at " + str(coords), SimUtils.MSG_MSG)
                self.ClearLoc(coords)
                return None
            else:
                SimUtils.PrintDebug("Adding a growth tile frontier location at " + str(coords), SimUtils.MSG_MSG)
                self.SetLoc(coords, TileAssembly.FRONTIER_LOC)
                return fLoc
        #print("fitting nbr set: " + str(fittingNbrSet))
        
    def IsSetOfGrowthTiles(self, tileTypes):
        for tt in tileTypes:
            if tt.startswith("growth") == False:
                return False
        return True
    
    def IsThreeContiguous(self, nbrSet):
        #if not len(nbrSet) == 3:
        #    print("not 3 contiguous: " + str(nbrSet))
            
        dirFromNbr = {(0,1):0, (1,1):1, (1,0):2, (1,-1):3, (0,-1):4, (-1,-1):5, (-1,0):6, (-1,1):7}
        dirs = []
        for nbr in nbrSet:
            dirs.append(dirFromNbr[nbr])
        for dir in dirs:
            if (dir+1)%8 in dirs and (dir+2)%8 in dirs:
                return True
        return False
    
    def IsTwoContiguous(self, nbrSet):
        if len(nbrSet) == 2:
            dirFromNbr = {(0,1):0, (1,1):1, (1,0):2, (1,-1):3, (0,-1):4, (-1,-1):5, (-1,0):6, (-1,1):7}
            dirs = []
            for nbr in nbrSet:
                dirs.append(dirFromNbr[nbr])
            for dir in dirs:
                if (dir+1)%8 in dirs:
                    return True
        return False
            
    def GetFittingTilesForLoc(self, coords):
        retSet = None
        count = 0
        numNbrs = 0
        nbrList = []
        matchSets = dict()
        for n in TileAssembly.NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrTile = self.GetLoc(nbrCoords, True)
            if not nbrTile is None:
                nbrList.append(n)
                oppDir = (count + 4) % 8
                matchSets[n] = nbrTile.type.matches[oppDir]
                numNbrs = numNbrs + 1                    
            count = count + 1
        
        if len(nbrList) > 3:
            bOk = True
            #print("Extra neighbors (" + str(len(nbrList)) + " for location " + str(coords))
            setsOf3 = set(itertools.combinations(matchSets.keys(), 3))
            singleMatches = []
            for so3 in setsOf3:
                l = list(so3)
                match = matchSets[l[0]].intersection(matchSets[l[1]].intersection(matchSets[l[2]]))
                if len(match) > 1:
                    #print("big match set for coords " + str(coords) + " and nbrSet " + str(so3) + ", len match: " + str(len(match)) + ", match set: " + str(match))
                    bOk = False
                elif len(match) == 1:
                    singleMatches.append(match)
                else:
                    print("Size 0 match set!")
            if len(singleMatches) > 1:
                #print(" There are " + str(len(singleMatches)) + " match sets with one entry: " + str(singleMatches))
                for i in range(len(singleMatches)-1):
                    if not singleMatches[i] == singleMatches[i+1]:
                        print("PROBLEM: Unequal size 1 match sets!")
            #if bOk == False:
                    #else:
                    #    print("They match")
            #    self.subsetOfNbrsToPlace.append(coords)
            
        for ms in matchSets.keys():
            if retSet is None:
                retSet = matchSets[ms]
            else:
                retSet = retSet.intersection(matchSets[ms])

        if not retSet is None:
            if len(retSet) == 1:
                if self.IsThreeContiguous(nbrList) == False:
                    tt = min(retSet)
                    okTiles = ["UOPPdsd","UOPPdbd","UOPPdDd","UOMMdsd","UOMMdbd","UOMMdDd"]
                    ok = False
                    for ot in okTiles:
                        if ot in tt:
                            if self.IsTwoContiguous(nbrList) == False:
                                print("Uh oh! Allowing tile placement without 2 or 3 contiguous neighbors at " + str(coords) + ", nbrs: " + str(nbrList))
                            self.twoNbrsToPlace.append(coords)
                            ok = True
                            break
                    if ok == False:
                        return None
                #print("numNbrs: " + str(numNbrs))
                if self.inputDict.has_key(min(retSet)):
                    large = set(nbrList)
                    small = self.inputDict[min(retSet)]
                    if not len(large) == len(small):
                        if len(small) > len(large):
                            temp = large
                            large = small
                            small = temp
                        #print(min(retSet) + " at " + str(coords) + " already in inputDict with different length input set. Large: " + str(len(large)) + ", small: " + str(len(small)))
                    #elif len(small) == len(large):
                        #if not small == large:
                            #print("same size but different: " + str(large) + " and " + str(small))
                    self.inputDict[min(retSet)] = set(nbrList)
                else:
                    self.inputDict[min(retSet)] = set(nbrList)
            elif self.IsSetOfGrowthTiles(retSet) == True:
                #print("numNbrs (growth): " + str(numNbrs))
                if self.IsTwoContiguous(nbrList) == False:
                    print("Uh oh! Allowing growth tile placement without 2 or 3 contiguous neighbors at " + str(coords) + ", nbrs: " + str(nbrList))
                self.twoNbrsToPlace.append(coords)
        return retSet
    
    def StepForward(self):
        #time.sleep(0.1)
        if self.currCacheLoc < len(self.cache) - 1:
            self.currCacheLoc = self.currCacheLoc + 1
            cacheLoc = self.cache[self.currCacheLoc]
            for at in cacheLoc.addedTiles:
                if self.SetLoc(cacheLoc.addedTiles[0].coords, cacheLoc.addedTiles[0].type, growIfNeeded=True) == True:
                    self.numTiles = self.numTiles + 1
                else:
                    print("ERROR in StepForward trying to use cached step")
            for rt in cacheLoc.removedTiles:
                self.RemoveTile(rt.coords)
            for afl in cacheLoc.addedFrontierLocs:
                self.AddToFrontierIfNecessary(afl.coords)
            for rfl in cacheLoc.removedFrontierLocs:
                self.frontier.remove(rfl)
            return (cacheLoc.addedTiles, cacheLoc.removedTiles, cacheLoc.addedFrontierLocs, cacheLoc.removedFrontierLocs)
        # Pick a spot from the frontier, add a tile
        elif len(self.frontier) == 0:
            SimUtils.globalApp.OutputMessage("The frontier is empty!")
            return None
        else:
            enabledFrontier = self.frontier
            if len(self.disabledFrontier) > 0:
                enabledFrontier = self.frontier - self.disabledFrontier
                if len(enabledFrontier) == 0:
                    SimUtils.globalApp.OutputMessage("There are no enabled frontier locations!")
                    return None
            frontierLoc = random.sample(enabledFrontier, 1)[0]
            if frontierLoc.isGrowthLoc == True:
                if self.growthPatternLoc >= len(self.growthPattern):
                    typeName = frontierLoc.GetRandomFittingTileTypeName()
                else:
                    pattern = self.growthPattern[self.growthPatternLoc]
                    matchingTypeNames = []
                    for ttName in frontierLoc.fittingTypeNames:
                        if pattern in ttName:
                            matchingTypeNames.append(ttName)
                    if len(matchingTypeNames) == 0:
                        SimUtils.globalApp.OutputMessage("ERROR: No fitting growth tile at " + str(frontierLoc.coords) + " matches pattern '" + str(pattern) + "'!  Available set was " + str(frontierLoc.fittingTypeNames))
                        self.growthPatternLoc = self.growthPatternLoc + 1
                        return None
                    else:
                        SimUtils.globalApp.OutputMessage("The growth tiles for location " + str(frontierLoc.coords) + " which match '" + str(pattern) + "' are these " + str(len(matchingTypeNames)) + ": " + str(matchingTypeNames))
                        typeName = random.sample(matchingTypeNames, 1)[0]
                        self.growthPatternLoc = self.growthPatternLoc + 1
                added = self.AddTile(self.tileTypes[typeName], frontierLoc.coords)
                if added[0] == True:
                    self.growthTileLocations.append(frontierLoc.coords)
                    newTile = Tile(self.tileTypes[typeName], frontierLoc.coords)
                    self.AddCacheEntry(addedTiles=[newTile], addedFrontierLocs=added[1], removedFrontierLocs=[frontierLoc])
                    SimUtils.globalApp.OutputMessage("Added a growth tile of type " + typeName + " to location " + str(frontierLoc.coords) + " from " + str(len(frontierLoc.fittingTypeNames)) + " choices " + str(frontierLoc.fittingTypeNames))
                    return ([newTile], [], added[1], [])
            else:
                typeName = frontierLoc.GetRandomFittingTileTypeName()
                added = self.AddTile(self.tileTypes[typeName], frontierLoc.coords)
                if added[0] == True:
                    newTile = Tile(self.tileTypes[typeName], frontierLoc.coords)
                    self.AddCacheEntry(addedTiles=[newTile], addedFrontierLocs=added[1], removedFrontierLocs=[frontierLoc])
                    SimUtils.PrintDebug("Added a tile of type " + typeName + " to location " + str(frontierLoc.coords), SimUtils.MSG_MSG)
                    return ([newTile], [], added[1], [])
            SimUtils.globalApp.OutputMessage("Error adding tile to location" + str(frontierLoc.coords))
            return None
        
    def AssemblyToString(self, tileList=None):
        assemblyString = "{"
        comma = ""
        count = 0
        if tileList is None:
            for x in range(self.minX, self.maxX+1):
                for y in range(self.minY, self.maxY+1):
                    loc = self.GetLoc((x,y))
                    if not loc is None:
                        count = count + 1
                        dash = loc.type.name.rfind('_')
                        nameNoOrient = loc.type.name[0:dash]
                        orient = loc.type.name[dash+1:]
                        assemblyString = assemblyString + comma + "{\"" + nameNoOrient + "\", {{" + str(loc.coords[0]) + ", " + str(loc.coords[1]) + "}, " + orient + "}}"
                        comma = ","
        else:
            for loc in tileList:
                count = count + 1
                dash = loc.type.name.rfind('_')
                nameNoOrient = loc.type.name[0:dash]
                orient = loc.type.name[dash+1:]
                assemblyString = assemblyString + comma + "{\"" + nameNoOrient + "\", {{" + str(loc.coords[0]) + ", " + str(loc.coords[1]) + "}, " + orient + "}}"
                comma = ","
        return assemblyString + "}"

    def LoadTileSystem(self, fileName):
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        #try:
        os.stat(fileName)
        rFile = open(fileName, "r")
        SimUtils.globalApp.OutputMessage("Using tile system definition file '" + fileName + "'")
        lines = rFile.readlines()
        rFile.close()
        
        self.systemDefFile = fileName
        self.tileSystemDef = ast.literal_eval(lines[0])
        self.tileSystemDef['MainFile'] = fileName
        self.tileSystemDef['DirName'] = os.path.dirname(fileName)
        
        print("Read tile system " + str(self.tileSystemDef))

        self.tileTypes = dict()
        self.seedPatch = []
        #self.growthTileTypes = []
        self.frontier = set()
        
        fileName = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['NeighborList']
        SimUtils.globalApp.OutputMessage("Reading neighbor list file '" + fileName + "'")
        os.stat(fileName)
        rFile = open(fileName, "r")
        lines = rFile.readlines()
        rFile.close()
                
        bHaveMarkings = False
        if "TileMarkingsList" in self.tileSystemDef.keys():
            fileName = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['TileMarkingsList']
            SimUtils.globalApp.OutputMessage("Reading tile markings list file '" + fileName + "'")
            os.stat(fileName)
            rFile = open(fileName, "r")
            tileMarkingDefs = rFile.readlines()
            rFile.close()
            bHaveMarkings = True
            strengthDict = dict() 
            
            if "MarkingStrengths" in self.tileSystemDef.keys():
                fileName = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['MarkingStrengths']
                SimUtils.globalApp.OutputMessage("Reading marking strengths file '" + fileName + "'")
                os.stat(fileName)
                rFile = open(fileName, "r")
                strengthLines = rFile.readlines()
                rFile.close()

                for strengthLine in strengthLines:
                    strengthLine = strengthLine.strip()
                    strength = strengthLine.split('=')[0]
                    markings = eval(strengthLine.split('=')[1])
                    for m in markings:
                        strengthDict[str(m)] = strength
            self.tileSystemDef['MarkingStrengthDict'] = strengthDict
                    
            if "MarkingMatchingList" in self.tileSystemDef.keys():
                fileName = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['MarkingMatchingList']
                SimUtils.globalApp.OutputMessage("Reading marking matching file '" + fileName + "'")
                os.stat(fileName)
                rFile = open(fileName, "r")
                matchings = rFile.readlines()
                rFile.close()
    
                matchingDict = dict()
                for match in matchings:
                    match = match.strip()
                    fields = match.split()
                    matchingDict[fields[0]] = fields[1]
                self.tileSystemDef['MarkingMatchingDict'] = matchingDict

            dirAbbrevNameArray = ("N", "NE", "E", "SE", "S", "SW", "W", "NW")
            tileMarkingsDict = dict()
            for tileDef in tileMarkingDefs:
                tileDef = tileDef.strip()
                fields = tileDef.split()
                #print(fields)
                tileName = fields[0]
    
                markings = []
                markingStr = ""
                count = 0
                for dir in dirAbbrevNameArray:
                    strength = 2
                    label = fields[count+1]
                    if count > 3:
                        label = matchingDict[label]
                    #labelAndDir = str(label) + dir
                    if (label in strengthDict.keys()):
                        strength = strengthDict[label]
                    elif (matchingDict[label] in strengthDict.keys()):
                        strength = strengthDict[matchingDict[label]]
                    #elif labelAndDir in strengthDict.keys():
                        #strength = strengthDict[labelAndDir]
                        #if str(label) + dir == "34W":
                        #    print("found 34W")
                        #print("for tile " + tileName + " found one: '" + str(label) + dir +"' with strength " + str(strength))
                    markings.append((label, strength))
                    markingStr = markingStr + dirAbbrevNameArray[count] + ":" + label  + "(" + str(strength) + ") "
                    count = count + 1
                tileMarkingsDict[tileName] = (markingStr, markings)
        
        #for i in range(len(lines)):
        i = 0
        while i < len(lines):
            name = lines[i].split()
            name = name[0]
            markings = name[1]
            i = i + 2
            
            dirs = []
            for j in range(8):
                #dir = lines[i].strip()
                i = i + 1

                matches = lines[i].strip()
                i = i + 2
                matches = matches.split()
                matchSet = set(matches)
                dirs.append(matchSet)

            i = i + 2
            #tileType = {"name":name, "matches":dirs}
            name = name.replace('"', '')
            img = "TileImage_" + name + ".tif"
            fullPath = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['TilePictureDir'] + "/" + img
            if os.path.isfile(fullPath) == True:
                tileType = RobinsonTiles.RobinsonTile(name, dirs, fullPath, self.tileWidths, self.tileHeights)
                if bHaveMarkings == True:
                    if name in tileMarkingsDict.keys():
                        tileType.markings = tileMarkingsDict[name]
                    else:
                        print("ERROR: no entry in tile markings list for " + name)
                else:
                    tileType.markings = None
                self.tileTypes[name] = tileType
            else:
                SimUtils.globalApp.OutputMessage("WARNING:  Missing image file: '" + img + "'")
                tileType = RobinsonTiles.RobinsonTile(name, dirs, None, self.tileWidths, self.tileHeights)
                if bHaveMarkings == True:
                    if name in tileMarkingsDict.keys():
                        tileType.markings = tileMarkingsDict[name]
                    else:
                        print("ERROR: no entry in tile markings list for " + name)
                else:
                    tileType.markings = None
                self.tileTypes[name] = tileType
            #print("tileType: " + name)
        SimUtils.globalApp.OutputMessage("Loaded tile set with size: " + str(len(self.tileTypes)))

        fileName = self.tileSystemDef['DirName'] + "/" + self.tileSystemDef['SeedPatch']
        SimUtils.globalApp.OutputMessage("Reading seed patch file '" + fileName + "'")

        midPt = self.LoadAssembly(fileName=fileName)
        return midPt
    
    def SaveSAAssembly(self, fileName, tdsFileName):
        wFile = open(fileName, "w")

        wFile.write(tdsFileName + "\n")
        wFile.write("Temperature=6\n")

        count = 0
        for x in range(self.minX, self.maxX+1):
            for y in range(self.minY, self.maxY+1):
                loc = self.GetLoc((x,y))
                if not loc is None:
                    wFile.write(loc.type.name + "-SEED " + str(x) + " " + str(y) + "\n")
                    #wFile.write(loc.type.name + " " + str(x) + " " + str(y) + "\n")
                    count = count + 1
        wFile.close()
        SimUtils.globalApp.OutputMessage("Saved assembly of " + str(count) + " tiles to file '" + fileName + "'")
    
    def AssemblyFromString(self, assemblyString):
        assemblyString = assemblyString.replace("{", "[")
        assemblyString = assemblyString.replace("}", "]")
        assemblyPreList = []
        if len(assemblyString) > 0:
            bException = True
            while bException == True:
                try:
                    assemblyPreList = eval(assemblyString)
                    bException = False
                except NameError as e:
                    errStr = str(e)
                    errArr = errStr.split("\'")
                    name = errArr[1]
                    assemblyString = assemblyString.replace(name, "\'" + name + "\'")
                    bException = True

        assemblyList = []
        for entry in assemblyPreList:
            name = entry[0]
            x = int(entry[1][0][0])
            y = int(entry[1][0][1])
            orient = entry[1][1]
            name = name + "_" + str(orient)
            assemblyList.append((name, (x,y)))
        return assemblyList
    
    def LoadAssembly(self, seedDefn=None, fileName=None):
        if not fileName is None:
            os.stat(fileName)
            rFile = open(fileName, "r")
            lines = rFile.readlines()
            rFile.close()
            seedDefn = ""
            for line in lines:
                seedDefn = seedDefn + line.strip()
            
        self.seedPatch = self.AssemblyFromString(seedDefn)
        self.ClearSpace()
        self.frontier = set()
        (aveX, aveY) = self.PlaceSeed()
        SimUtils.globalApp.OutputMessage("Loaded assembly with " + str(len(self.seedPatch)) + " tiles")
        return (aveX, aveY)

    def AddTile(self, ttype, coords):
        retVal = TileAssembly.Assembly.AddTile(self, ttype, coords)
        self.tileTypesUsed.add(ttype.name)
        return retVal
        
    def PrintStats(self):
        SimUtils.globalApp.OutputMessage("Tile set size: " + str(len(self.tileTypes)))
        SimUtils.globalApp.OutputMessage("Tile types used in this simulation: " + str(len(self.tileTypesUsed)))
        usedTilesIn = open("usedTileTypes.txt", "r")
        usedSet = set()
        lines = usedTilesIn.readlines()
        for line in lines:
            typeName = line.strip()
            usedSet.add(typeName)
        usedTilesIn.close()
        usedSet = self.tileTypesUsed | usedSet
        SimUtils.globalApp.OutputMessage("Combined used tile types: " + str(len(usedSet)))
        usedTilesOut = open("usedTileTypes.txt", "w")
        for tt in usedSet:
            usedTilesOut.write(tt + "\n")
        usedTilesOut.close()
        for tt in self.inputDict:
            print(tt + ": " + str(self.inputDict[tt]))
        #=======================================================================
        # for tt in usedSet:
        #     for d in range(8):
        #         dSet = self.tileTypes[tt].matches[d]
        #         for tt2 in dSet:
        #             oppDir = (d + 4) % 8
        #             oppSet = self.tileTypes[tt2].matches[oppDir]
        #             if not dSet == oppSet:
        #                 print("mismatch: " + str(dSet) + "  " + str(oppSet))
        #=======================================================================
        
        badTilesIn = open("badTiles.txt", "r")
        badSet = set()
        lines = badTilesIn.readlines()
        for line in lines:
            typeName = line.strip()
            badSet.add(typeName)
        badTilesIn.close()
        #print("num bad tiles: " + str(len(badSet)))

        truncUsedSet = set()
        for tt in usedSet:
            trunc = tt.rsplit("_",1)[0]
            truncUsedSet.add(trunc)
        #print("truncated used set: " + str(truncUsedSet))
        
        badUsed = badSet.intersection(truncUsedSet)
        print("badUsed: " + str(badUsed))
        
#===============================================================================
# for each tile t:
#     for each direction d:
#         for each tile t2 that matches t on d:
#             for each tile t3 that matches t and t2:
#                 for each tile t4 that matches t and t3:
#                     how many tiles can bind to t2 and t3 and t4?
#===============================================================================

        dirFromNbr = {(0,1):0, (1,1):1, (1,0):2, (1,-1):3, (0,-1):4, (-1,-1):5, (-1,0):6, (-1,1):7}

        #MJP: redo this to start from each tile in usedSet and form a ring of 3, with it as the first
        # - for each set of tiles that could form the ring with it, do they force a tile inside the ring?
        
        targetSetStats = dict()

        # target: location being tested for forcing
        # nbr1, nbr2, nbr3:  ring of nbrs of target in cw order
        dirNum = 0
        for nbr1 in TileAssembly.NBRS:
            nbr1SideWithTarget = (dirNum + 4) % 8
            
            nbr2 = TileAssembly.NBRS[(dirNum + 1) % 8]            
            nbr1DiffNbr2 = (nbr2[0] - nbr1[0], nbr2[1] - nbr1[1])
            nbr1SideWithNbr2 = dirFromNbr[nbr1DiffNbr2]
            
            nbr2SideWithTarget = ((dirNum + 1) + 4) % 8
            
            nbr3 = TileAssembly.NBRS[(dirNum + 2) % 8]            
            nbr2DiffNbr3 = (nbr3[0] - nbr2[0], nbr3[1] - nbr2[1])
            nbr2SideWithNbr3 = dirFromNbr[nbr2DiffNbr3]
            
            nbr3SideWithTarget = ((dirNum + 2) + 4) % 8
            
            for nbr1TTName in usedSet:
                nbr1TT = self.tileTypes[nbr1TTName]
                
                nbr1SetWithTarget = nbr1TT.matches[nbr1SideWithTarget].intersection(usedSet)
                nbr1SetWithNbr2 = nbr1TT.matches[nbr1SideWithNbr2].intersection(usedSet)
                
                for nbr2TTName in nbr1SetWithNbr2:
                    nbr2TT = self.tileTypes[nbr2TTName]
                    
                    nbr2SetWithTarget = nbr2TT.matches[nbr2SideWithTarget].intersection(usedSet)
                    nbr2SetWithNbr3 = nbr2TT.matches[nbr2SideWithNbr3].intersection(usedSet)
                    
                    for nbr3TTName in nbr2SetWithNbr3:
                        nbr3TT = self.tileTypes[nbr3TTName]
                        
                        nbr3SetWithTarget = nbr3TT.matches[nbr3SideWithTarget].intersection(usedSet)
                        
                        targetSet = nbr1SetWithTarget.intersection(nbr2SetWithTarget.intersection(nbr3SetWithTarget))
                        
                        lenTargetSet = len(targetSet)
                        if lenTargetSet in targetSetStats.keys():
                            targetSetStats[lenTargetSet] = targetSetStats[lenTargetSet] + 1
                        else:
                            targetSetStats[lenTargetSet] = 1
            
            dirNum = dirNum + 1
            
            print("Target set stats: " + str(targetSetStats))
        print("Done with stats")

        #=======================================================================
        # for t in usedSet:
        #     tt = self.tileTypes[t]
        #     dir1 = 0
        #     for nbr1 in TileAssembly.NBRS:
        #         tNbrsDir1 = tt.matches[dir1].intersection(usedSet)
        #         
        #         dir2 = (dir1 + 1) % 8
        #         nbr2 = TileAssembly.NBRS[dir2]
        #         nbrDiff12 = (nbr2[0]-nbr1[0], nbr2[1]-nbr1[1])
        #         tNbrsDir2 = tt.matches[dir2].intersection(usedSet)
        #         
        #         dir3 = (dir1 + 2) % 8
        #         nbr3 = TileAssembly.NBRS[dir3]
        #         nbrDiff23 = (nbr3[0]-nbr2[0], nbr3[1]-nbr2[1])
        #         tNbrsDir3 = tt.matches[dir3].intersection(usedSet)
        #         
        #         for ttNbr1 in tNbrsDir1:
        #             nbr1NbrSet = self.tileTypes[ttNbr1].matches[dirFromNbr[nbrDiff12]]
        #             
        #             for ttNbr2 in nbr1NbrSet:
        #                 nbr2NbrSet = self.tileTypes[ttNbr2].matches[dirFromNbr[nbrDiff23]]
        #         
        #         dir = dir + 1
        #=======================================================================
